/*
 *  Optimized version of PerspectiveTransform.js
 *  by Edan Kwan
 *  website: http://www.edankwan.com/
 *  twitter: https://twitter.com/#!/edankwan
 *  Lab: www.edankwan.com/lab
 *
 *  The original PerspectiveTransform.js is created by  Israel Pastrana
 *  http://www.is-real.net/experiments/css3/wonder-webkit/js/real/display/PerspectiveTransform.js
 *
 *  Matrix Libraries from a Java port of JAMA: A Java Matrix Package, http://math.nist.gov/javanumerics/jama/
 *  Developed by Dr Peter Coxhead: http://www.cs.bham.ac.uk/~pxc/
 *  Available here: http://www.cs.bham.ac.uk/~pxc/js/
 *
 *  I simply removed some irrelevant variables and functions and merge everything into a smaller function. I also added some error checking functions and bug fixing things.
 */

(function (root, factory) {
  if(typeof define === "function" && define.amd) {
    define(factory);
  } else {
    root.PerspectiveTransform = factory();
  }
}(this, function(){
    var transformStyleName;
    var transformOriginDomStyleName;

    (function(){
        var testStyle = document.createElement('div').style;
        var stylePrefix =
            'webkitTransform' in testStyle ? 'webkit' :
            'MozTransform' in testStyle ? 'Moz' :
            'msTransform' in testStyle ? 'ms' :
            '';
        transformStyleName = stylePrefix + (stylePrefix.length>0?'Transform':'transform');
        transformOriginDomStyleName = '-'+ stylePrefix.toLowerCase()+'-transform-origin';
    })();


    // Check the distances between each points and if there is some points with the distance lequal to or less than 1 pixel, then return true. Otherwise return false;
    function hasDistancesError(){
        var lenX = this.topLeft.x - this.topRight.x;
        var lenY = this.topLeft.y - this.topRight.y;
        if(Math.sqrt(lenX * lenX +  lenY * lenY)<=1) return true;
        lenX = this.bottomLeft.x - this.bottomRight.x;
        lenY = this.bottomLeft.y - this.bottomRight.y;
        if(Math.sqrt(lenX * lenX +  lenY * lenY)<=1) return true;
        lenX = this.topLeft.x - this.bottomLeft.x;
        lenY = this.topLeft.y - this.bottomLeft.y;
        if(Math.sqrt(lenX * lenX +  lenY * lenY)<=1) return true;
        lenX = this.topRight.x - this.bottomRight.x;
        lenY = this.topRight.y - this.bottomRight.y;
        if( Math.sqrt(lenX * lenX +  lenY * lenY)<=1) return true;
        lenX = this.topLeft.x - this.bottomRight.x;
        lenY = this.topLeft.y - this.bottomRight.y;
        if( Math.sqrt(lenX * lenX +  lenY * lenY)<=1) return true;
        lenX = this.topRight.x - this.bottomLeft.x;
        lenY = this.topRight.y - this.bottomLeft.y;
        if( Math.sqrt(lenX * lenX +  lenY * lenY)<=1) return true;

        return false;
    }

    // Get the determinant of given 3 points
    function getDeterminant(p0, p1, p2){
        return p0.x * p1.y + p1.x * p2.y + p2.x * p0.y - p0.y * p1.x - p1.y * p2.x - p2.y * p0.x;
    }

    // Return true if it is a concave polygon or if it is backfacing when the useBackFacing property is false. Otehrwise return true;
    function hasPolyonError(){
        var det1 = getDeterminant(this.topLeft, this.topRight, this.bottomRight);
        var det2 = getDeterminant(this.bottomRight, this.bottomLeft, this.topLeft);
        if(this.useBackFacing){
            if(det1*det2<=0) return true;
        }else{
            if(det1<=0||det2<=0) return true;
        }
        var det1 = getDeterminant(this.topRight, this.bottomRight, this.bottomLeft);
        var det2 = getDeterminant(this.bottomLeft, this.topLeft, this.topRight);
        if(this.useBackFacing){
            if(det1*det2<=0) return true;
        }else{
            if(det1<=0||det2<=0) return true;
        }
        return false;
    }


    function PerspectiveTransform(element, width, height, useBackFacing){
        this.element = element;
        this.style = element.style;
        this.computedStyle = window.getComputedStyle(element);
        this.width = width;
        this.height = height;
        this.useBackFacing = !!useBackFacing;

        this.topLeft = {x: 0, y: 0};
        this.topRight = {x: width, y: 0};
        this.bottomLeft = {x: 0, y: height};
        this.bottomRight = {x: width, y: height};
    }


    PerspectiveTransform.translate = function(points, translateX, translateY){
        var point;
        for (var i=0; i < points.length; i++){
            point = points[i];
            point[0] += translateX;
            point[1] += translateY;
        }

        return points
    };


    PerspectiveTransform.matrix3d = function(sourcePoints, targetPoints) {
        for (var a = [], b = [], i = 0, n = sourcePoints.length; i < n; ++i) {
            var s = sourcePoints[i], t = targetPoints[i];
            a.push([s[0], s[1], 1, 0, 0, 0, -s[0] * t[0], -s[1] * t[0]]), b.push(t[0]);
            a.push([0, 0, 0, s[0], s[1], 1, -s[0] * t[1], -s[1] * t[1]]), b.push(t[1]);
        }

        var X = linearEquationSolver(a, b, true), matrix = [
            X[0], X[3], 0, X[6],
            X[1], X[4], 0, X[7],
               0,    0, 1,    0,
            X[2], X[5], 0,    1
        ].map(function(x) {
            return x.toFixed(9)
        });

        return matrix;
    }

    PerspectiveTransform.prototype.update = function() {
        var width = this.width;
        var height = this.height;

        //  get the offset from the transfrom origin of the element
        var offsetX = 0;
        var offsetY = 0;
        var offset = this.computedStyle.getPropertyValue(transformOriginDomStyleName);
        if(offset.indexOf('px')>-1){
            offset = offset.split('px');
            offsetX = parseFloat(offset[0]);
            offsetY = parseFloat(offset[1]);
        }else if(offset.indexOf('%')>-1){
            offset = offset.split('%');
            offsetX = parseFloat(offset[0]) * width / 100;
            offsetY = parseFloat(offset[1]) * height / 100;
        }

        var sourcePoints = PerspectiveTransform.translate([
            [0,0],
            [width, 0],
            [0, height],
            [width, height]
        ],  -offsetX, -offsetY);

        var targetPoints = PerspectiveTransform.translate([
            [this.topLeft.x, this.topLeft.y],
            [this.topRight.x, this.topRight.y],
            [this.bottomLeft.x, this.bottomLeft.y],
            [this.bottomRight.x, this.bottomRight.y]
        ], -offsetX, -offsetY);


        var matrix3D = PerspectiveTransform.matrix3d(sourcePoints, targetPoints);
        this.style[transformStyleName] = "matrix3d(" + matrix3D + ")";
    }

    PerspectiveTransform.prototype.checkError = function(){
        if(hasDistancesError.apply(this)) return 1; // Points are too close to each other.
        if(hasPolyonError.apply(this)) return 2; // Concave or backfacing if the useBackFacing property is false
        return 0; // no error
    }

    return PerspectiveTransform;
}));
